(function(window) {
var Generator = function(tileDefs, tileSizeFactor, rectAddedCallback, rectRemovedCallback)
{
	this.initialize(tileDefs, tileSizeFactor, rectAddedCallback, rectRemovedCallback);
}
var p = Generator.prototype;

p.initialize = function(tileDefs, tileSizeFactor, rectAddedCallback, rectRemovedCallback)
{
	this.tileDefs = tileDefs;
	this.tileSizeFactor = tileSizeFactor;//perc of total viewport area
	this.rectAddedCallback = rectAddedCallback;
	this.rectRemovedCallback = rectRemovedCallback;
	//objects for pooling
	this.p0 = new Point();
	this.p1 = new Point();
	// Tile Z-form:
	//  ##
	//   ##
	// Smallest square pattern of this tile is 4x4
	//Create map for looking up offset to tile reg point for each cell in this 4x4 pattern
	this.tilePatternSize = 4;
	this.tileSize = 4;//nr of gridcells in tile
	this.tileOffsets = [
		vec2.fromValues(0,0),
		vec2.fromValues(-1,0),
		vec2.fromValues(-1,-1),
		vec2.fromValues(-2,-1),
		
		vec2.fromValues(-1,0),
		vec2.fromValues(-1,-1),
		vec2.fromValues(-2,-1),
		vec2.fromValues(0,0),
		
		vec2.fromValues(-1,-1),
		vec2.fromValues(-2,-1),
		vec2.fromValues(0,0),
		vec2.fromValues(-1,0),
		
		vec2.fromValues(-2,-1),
		vec2.fromValues(0,0),
		vec2.fromValues(-1,0),
		vec2.fromValues(-1,-1)
	];
	this.tileSymmetries = [0,Math.PI];
	this.tileCenter = vec2.fromValues(1.5,1);//in cells
	//
	this.idcount = 0;
}


p.generate = function(viewport)
{
	this.tiles = {};//tiles are stored by hash. A tile is a collection of rects
	//determine gridcell width and height based on viewport w & h & tileSizeFactor
	var area = (viewport.width*viewport.height) * this.tileSizeFactor / this.tileSize;
	this.cellSize = Math.sqrt(area);
	//log(viewport.width, viewport.height, viewport.width*viewport.height, this.cellSize);
	this.setViewport(viewport);
}

p.setViewport = function(r)
{
	this.viewport = r;
	//Check which tiles are now inside/outside viewport
	var s = this.cellSize;
	var gl = Math.floor(r.x/s);
	var gr = Math.ceil((r.x+r.width)/s);
	var gt = Math.floor(r.y/s);
	var gb = Math.ceil((r.y+r.height)/s);
	//log("Gen.setViewport", r.width,r.height,s,gl,gr,gt,gb);
	var hashesToKeep = {};
	for (var y=gt;y<gb;y++)
	{
		for (var x=gl;x<gr;x++)
		{
			var tile = this.getTile(x,y);
			if (!tile)
			{
				tile = this.createTileAt(x,y);
			}
			var hash = this.getTileHash(x,y);
			hashesToKeep[hash] = true;
		}
	}
	//now check which tiles to remove
	for (hash in this.tiles)
	{
		if (!hashesToKeep[hash]) this.removeTile(this.tiles[hash], hash);
	}
	//
}

p.createTileAt = function(gx,gy)
{
	var s = this.cellSize;
	var vc = this.tileCenter;//vector of tile center (normalized)
	//
	var tile = [];//collection of rects
	var td = RandomUtil.pick(this.tileDefs);
	//pick rotational symmetry to apply
	var angle = RandomUtil.pick(this.tileSymmetries);
	//correct gx and gy so they point to tile origin on grid
	var vreg = this.getTileReg(gx,gy);
	gx = vreg[0];
	gy = vreg[1];
	var ox = gx*s;//origin of tile to place in world coords
	var oy = gy*s;
	for (var i=0;i<td.length;i+=4)
	{
		//apply angle: rotate rect TL & BR around tile center
		var w = td[i+2];
		var h = td[i+3];
		//find TL
		var cx = td[i] - vc[0];//offset of TL of rect to center of tile
		var cy = td[i+1] - vc[1];
		var a = Math.atan2(cy,cx);
		var len = Math.sqrt(cx*cx+cy*cy);
		a += angle;
		var rl = vc[0] + Math.cos(a)*len;
		var rt = vc[1] + Math.sin(a)*len;
		//find BR
		cx = td[i]+w - vc[0];
		cy = td[i+1]+h - vc[1];
		a = Math.atan2(cy,cx);
		len = Math.sqrt(cx*cx+cy*cy);
		a += angle;
		var rr = vc[0] + Math.cos(a)*len;
		var rb = vc[1] + Math.sin(a)*len;
		//determine rect from T,L,B and R
		w = Math.abs(rl-rr);
		var rx = Math.min(rl,rr);
		h = Math.abs(rt-rb);
		var ry = Math.min(rt,rb);
		var r = new Rectangle(Math.floor(ox+s*rx), Math.floor(oy+s*ry), Math.floor(s*w), Math.floor(s*h));
		//add id and store rect
		r.id = "r"+(this.idcount++);
		//log(r.id, r.toString());
		tile.push(r);
		this.rectAddedCallback(r);
	}
	this.setTile(gx,gy, tile);
	return tile;
}

p.removeTile = function(tile, hash)
{
	for (var i=0;i<tile.length;i++)
	{
		var r = tile[i];
		this.rectRemovedCallback(r);
	}
	delete this.tiles[hash];
}



p.getTileReg = function(gx,gy)
{
	//gx & gy should fulfill gy = -gx+4k
	//so gy+gx = 4k, with k int
	var s = this.tilePatternSize;
	var offset = this.tileOffsets[((gy%s+s)%s)*s + ((gx%s+s)%s)];
	var v = vec2.fromValues(gx,gy);
	return vec2.add(v, offset, v);
}
p.getTileHash = function(gx,gy)
{
	//gx & gy should fulfill gy = -gx+4k
	//so gy+gx = 4k, with k int
	//if (gy+gx)/4 not integer, then find key coord of tile that contains gx,gy
	var v = this.getTileReg(gx,gy);
	return "t"+v[0]+"_"+v[1];
}
p.getTile = function(gx,gy)
{
	var hash = this.getTileHash(gx,gy);
	return this.tiles[hash];
}
p.setTile = function(gx,gy, tile)
{
	var hash = this.getTileHash(gx,gy);
	this.tiles[hash] = tile;
}


window.Generator = Generator;
}(window));